/**
 * Copyright (C) 2005-2006 Funambol
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.email.cache;

import java.io.Serializable;
import java.security.Principal;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.logging.*;

import com.funambol.framework.core.*;
import com.funambol.framework.engine.*;
import com.funambol.framework.engine.source.*;
import com.funambol.framework.security.*;
import com.funambol.framework.tools.beans.*;
import com.funambol.framework.filter.FilterClause;

import com.funambol.email.admin.EmailConnectorConfig;
import com.funambol.email.exception.EmailConfigException;
import com.funambol.email.util.Def;


/**
 * This class represents a SyncSource that read and store the state of the items
 * in a support table.
 *
 * @version $Id: CrcCachedSyncSource.java,v 1.3 2006/02/15 16:41:21 gbmiglia Exp $
 *
 */
public abstract class CrcCachedSyncSource
        extends AbstractSyncSource
        implements SyncSource, Serializable, LazyInitBean {

    // --------------------------------------------------------------- Constants

    private Logger log = Logger.getLogger(Def.LOGGER_NAME);

    private String jndiDataSourceName = null;

    // ------------------------------------------------------------ Private Data

    /**    */
    private CrcItemStateDAO crcItemStateDao = null;

    /**    */
    private Principal principal          = null;
    /**    */
    private int       syncMode;

    /**
     *
     */
    private CrcSyncItemInfo[] localItems    = null;
    private CrcSyncItemInfo[] serverItems   = null;

    /*
     *
     */
    private ArrayList newItemKeys     = null;
    private ArrayList deletedItemKeys = null;
    private ArrayList updatedItemKeys = null;

    /**
     * keep the items id and the result of send operation
     *
     */
    public ArrayList itemsSent = new ArrayList();



    // ---------------------------------------------------------- Public Methods

    /**
     * Init the syncSource
     *
     * @throws BeanInitializationException
     */
    public void init() throws BeanInitializationException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Init CrcCachedSyncSource with sourceUri: " + sourceURI);
        }

        try {
          jndiDataSourceName  = EmailConnectorConfig.getConfigInstance().getDataSource();
        } catch (EmailConfigException e) {
            log.severe("Error Getting jndi datasource Parameters:: " + e);
            log.throwing(getClass().getName(), "checkMailCredentials", e);
        }

        crcItemStateDao = new CrcItemStateDAO(jndiDataSourceName);
        try {
            crcItemStateDao.init();
        } catch (Exception e) {
            throw new BeanInitializationException("Error initializing CrcItemStateDAO", e);
        }
    }

    /**
     * BeginSync
     *
     * @param principal Principal
     * @param syncMode int
     * @param filterClause FilterClause
     * @throws SyncSourceException
     */
    public void beginSync(Principal principal,
                          int syncMode,
                          FilterClause filterClause)
        throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource beginSync [" + sourceURI + "] ");
        }

        this.principal = principal;
        this.syncMode  = syncMode;

        String principalId = ((Sync4jPrincipal) this.principal).getId();

        try {

            if ((syncMode != AlertCode.SLOW) &&
                (syncMode != AlertCode.REFRESH_FROM_SERVER)) {

                // get all items from EIS server
                this.serverItems = getAllSyncItemInfo();

                // get all items from cache
                this.localItems  = crcItemStateDao.getLocalItems(this.sourceURI,
                                                                 principalId);

                // create the array with the new, updated, deleted items
                this.newItemKeys     = setNewItemKeys();
                this.updatedItemKeys = setUpdatedItemIds();
                this.deletedItemKeys = setDeletedItemKeys();

            }

        } catch (Exception e) {
            throw new SyncSourceException("Error accessing items: " +
                                           e.getMessage(), e);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Crc Cached NEW");
            if (this.newItemKeys != null){
               log.finest("Crc Cached NEW num:" + this.newItemKeys.size());
            } else {
               log.finest("Crc Cached NO NEW ITEMS");
            }

            log.finest("Crc Cached UPDATED");
            if (this.updatedItemKeys != null){
               log.finest("Crc Cached UPDATED num:" + this.updatedItemKeys.size());
            } else {
               log.finest("Crc Cached NO UPDATED ITEMS");
            }

            log.finest("Crc Cached DELETED");
            if (this.deletedItemKeys != null){
               log.finest("Crc Cached DELETED num:" + this.deletedItemKeys.size());
            } else {
               log.finest("Crc Cached NO DELETED ITEMS");
            }
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource beginSync done [" + sourceURI + "]");
        }

    }

    /**
     * EndSync
     *
     * @todo:optimize the performance,
     * by read only update / add items
     * to setting lastModified
     * with addItems[exist], deleteItems[exist],
     * updateItems[implements] array
     *
     *
     * @throws SyncSourceException
     */
    public void endSync() throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource endSync [" + sourceURI + "]"
                       + " principal " + this.principal );
        }

        CrcSyncItemInfo[] serverItems = null;

        String principalId = ((Sync4jPrincipal) this.principal).getId();

        // reload cache
        try {
            serverItems  = getAllSyncItemInfo();

            if (serverItems != null){
               if (log.isLoggable(Level.FINEST)) {
                  log.finest("CrcCachedSyncSource server Items num: " + serverItems.length);
               }
            }
            crcItemStateDao.updateLocalItems (serverItems ,
                                              this.sourceURI ,
                                              principalId);

        } catch (Exception e) {
            throw new SyncSourceException("Error caching items:" +
                                          e.getMessage(), e);
        }



        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource endSync done [" + sourceURI + "]");
        }

    }

    /**
     *
     * @see SyncSource
     *
     * @param since Timestamp
     * @param to Timestamp
     * @return SyncItemKey[]
     * @throws SyncSourceException
     */
    public SyncItemKey[] getUpdatedSyncItemKeys(Timestamp since, Timestamp to)
            throws SyncSourceException {
        return (SyncItemKey[])updatedItemKeys.toArray(new SyncItemKey[0]) ;
    }

    /**
     *
     * @see SyncSource
     *
     * @param since Timestamp
     * @param to Timestamp
     * @return SyncItemKey[]
     * @throws SyncSourceException
     */
    public SyncItemKey[] getNewSyncItemKeys(Timestamp since, Timestamp to)
            throws SyncSourceException {
        return (SyncItemKey[])newItemKeys.toArray(new SyncItemKey[0]);
    }

    /**
     *
     * @see SyncSource
     *
     * @param since Timestamp
     * @param to Timestamp
     * @return SyncItemKey[]
     * @throws SyncSourceException
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Timestamp since, Timestamp to)
            throws SyncSourceException {
        return (SyncItemKey[])deletedItemKeys.toArray(new SyncItemKey[0]) ;
    }



    /**
     *
     * @return CrcSyncItemInfo[]
     * @throws SyncSourceException
     */
    public abstract CrcSyncItemInfo[] getAllSyncItemInfo()
            throws  SyncSourceException;


    /**
     *
     *
     * @param syncItemKey SyncItemKey
     * @return char
     * @throws SyncSourceException
     */
    public char getSyncItemStateFromId(SyncItemKey syncItemKey)
      throws SyncSourceException {

        char status       = SyncItemState.UNKNOWN;
        boolean isNew     = false;
        boolean isUpdated = false;
        boolean isDeleted = false;

        String id = syncItemKey.getKeyAsString();

        //
        if ((this.syncMode == AlertCode.SLOW) ||
            (this.syncMode == AlertCode.REFRESH_FROM_SERVER)) {
           return SyncItemState.NEW;
        }

        // search in new items list
        //SyncItem tmp = null;
        SyncItemKey tmp = null;
        for (int i=0; i<newItemKeys.size(); i++){
            tmp = (SyncItemKey)newItemKeys.get(i);
            if (tmp.getKeyAsString().equals(id)) {
                isNew  = true;
                status = SyncItemState.NEW;
                break;
            }
        }

        if (!isNew){

           // search in updated items list
           for (int i=0; i<updatedItemKeys.size(); i++){
              tmp = (SyncItemKey)updatedItemKeys.get(i);
              if (tmp.getKeyAsString().equals(id)) {
                 isUpdated = true;
                 status    = SyncItemState.UPDATED;
                 break;
              }
           }


           if (!isUpdated){
              // search in deleted items list
              for (int i=0; i<deletedItemKeys.size(); i++){
                 tmp = (SyncItemKey)deletedItemKeys.get(i);
                 if (tmp.getKeyAsString().equals(id)) {
                    isDeleted = true;
                    status    = SyncItemState.DELETED;
                    break;
                 }
              }

              if (!isDeleted){
                 status    = SyncItemState.SYNCHRONIZED;
              }

           }

        }

        return status;

    }

    /**
     *
     *
     * @param id String
     * @throws SyncSourceException
     */
    public void insertSentItem(String id) throws SyncSourceException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("CrcCachedSyncSource insertSentItem ");
        }

        String principalId = ((Sync4jPrincipal) this.principal).getId();

        CrcSyncItemInfo[] sentItems = new CrcSyncItemInfo[1];

        CrcSyncItemInfo sentItem = new CrcSyncItemInfo();
        sentItem.setKey(new SyncItemKey(id));
        sentItem.setLastCrc(1);

        sentItems[0] = sentItem;

        try {
           crcItemStateDao.addLocalItems(sentItems, this.sourceURI, principalId);
        } catch (Exception e) {
            throw new SyncSourceException("Error caching sent items:" +
                                          e.getMessage(), e);
        }

    }


   // ---------------------------------------------------------- Private Methods

    /**
     * Creates new items id array
     *
     * @return new items id array
     */
    private ArrayList setNewItemKeys() {

        ArrayList  newRows         = null ;
        CrcSyncItemInfo serverItem  = null ;

        newRows = new ArrayList();

        int l = this.serverItems.length;

        for (int i = 0; i < l; i++) {

            serverItem = this.serverItems[i];

            if (!isItemInArray(this.localItems, serverItem)) {
                newRows.add(serverItem.getKey());
            }

        }
        return newRows;
    }

    /**
     * Creates deleted items id array
     *
     * @return deleted items id array
     */
    private ArrayList setDeletedItemKeys() {

        CrcSyncItemInfo localItem = null ;
        ArrayList deleteRows      = new ArrayList();

        for (int i=0, l = this.localItems.length; i < l; i++) {
           localItem = localItems[i];
           if (!this.isItemInArray(this.serverItems, localItem)) {
               deleteRows.add(localItem.getKey());
           }
        }

        return deleteRows;
    }

    /**
     * Creates updated items id array
     *
     * @return updated items id array
     */
    private ArrayList setUpdatedItemIds() {

        ArrayList updateRows      = new ArrayList();

        CrcSyncItemInfo serverItem   = null ;
        CrcSyncItemInfo localItem    = null ;

        long serverItemLastCrc ;
        long localItemLastCrc  ;

        for (int i = 0, l = this.serverItems.length; i < l; i++) {

            serverItem     = this.serverItems[i];
            localItem      = this.getItemById(this.localItems,
                                              serverItem.getKey().getKeyAsString()) ;

            if (localItem != null) {
                 serverItemLastCrc   = serverItem.getLastCrc () ;
                 localItemLastCrc    = localItem.getLastCrc  () ;

                 if (serverItemLastCrc != localItemLastCrc) {
                    updateRows.add(serverItem.getKey());
                 }
             }
        }

        //return (SyncItemKey[])updateRows.toArray(new SyncItemKey[0]) ;
        return updateRows ;

    }


    /**
     * Checks if item id is included in ids from array of items
     *
     * @param items items array
     * @param id id to find
     * @return <p>true</p> if the item is included
     *         <p>false</p> if the item is not included
     **/
    private boolean isIdInArray(CrcSyncItemInfo[] items,
                                String id) {

        boolean isFind = false;
        int l = items.length;

        for (int i = 0; i < l; i++) {

            if (log.isLoggable(Level.FINEST)) {
               log.finest("isIdInArray id 1: " + items[i].getKey().getKeyAsString() +
                          " id 2: " + id );
            }

            if (items[i].getKey().getKeyAsString().equals(id)) {
                isFind = true;
                break;
            }
        }

        return isFind;
    }

    /**
     * Checks if item is included in array of items
     *
     * @param items items array
     * @param item item to find
     * @return <p>true</p> if the item is included
     *         <p>false</p> if the item is not included
     **/
    private boolean isItemInArray(CrcSyncItemInfo[] items,
                                  CrcSyncItemInfo   item) {

        return isIdInArray(items, item.getKey().getKeyAsString());

    }

    /**
     * Gets item by id
     *
     *
     * @param items CrcSyncItemInfo[]
     * @param id String
     * @return <p>item</p> if found
     *         <p>null</p> if not found
     */
    private CrcSyncItemInfo getItemById(CrcSyncItemInfo[] items, String id) {
        CrcSyncItemInfo item = null;
        for (int i = 0, l = items.length; i < l; i++) {
            if (items[i].getKey().getKeyAsString().equals(id)) {
                item = items[i];
                break;
            }
        }
        return item;
    }


    /**
     *
     *
     * @param keys SyncItemKey[]
     * @param state char
     * @return SyncItem[]
     * @throws SyncSourceException
     */
    /*
    private SyncItem[] getSyncItems(SyncItemKey[] keys, char state) throws SyncSourceException  {

        SyncItem item = null;

        int num = keys.length;
        SyncItem[] items = new SyncItem[num];

        for (int i = 0; i < num; i++ ){
           item = getSyncItemFromId(keys[i]);
           item.setState(state);
           items[i] = item;
        }

        return items;
    }
    */

}
